from ormm.mathprog.milp import \
    resource_allocation, print_sol, sensitivity_analysis, blending

from ormm.mathprog.scheduling import scheduling

__all__ = ["resource_allocation", "print_sol",
           "sensitivity_analysis", "blending", "scheduling"]



"""
"""

import pyomo.environ as pyo


def scheduling(prob_class, linear=False, **kwargs):
    """
    Calls factory methods for different scheduling problems.

    The `prob_class` parameter allows the user to choose from different
    types of problem classes, which in turn have different model structures.

    The valid choices are:

    - :py:obj:`employee`:  A simple employee scheduling problem to
      minimize the number of workers employed to meet period requirements.
      Currently assumes that a worker works their periods in a row
      (determined by `ShiftLenth` parameter).

    - :py:obj:`rental`:  A type of scheduling problem where there are different
      plans (with different durations & costs), and the goal is to minimize the
      total cost of the plans purchased while meeting the period requirements
      (covering constraints).

    - :py:obj`agg_planning`: A planning problem where the decision variables
      are how much to produce during each period, to minimize production and
      holding costs while satisfying demand.

    More details on these model classes can be found in the Notes section here,
    as well as the corresponding section of the :ref:`math_prog`.

    Parameters
    ----------
    prob_class : :py:obj:`str`, optional
        Choice of "employee", "rental", or "agg_planning"
        to return different scheduling models.
    linear : :py:obj:`bool`, optional
        Determines whether decision variables will be
        Reals (True) or Integer (False).
    **kwargs
        Passed into Pyomo Abstract Model's `create_instance`
        to return Pyomo Concrete Model instead.

    Raises
    ------
    TypeError
        Raised if invalid argument value is given for `prob_class`.

    Notes
    -----
    The employee model minimizes workers hired with covering constraints:

    .. math::

       \\text{Min}  \\sum_{p \\in P} X_p

       \\text{s.t.} \\sum_{p - (L - 1)}^p X_p \\geq R_p
           \\quad \\forall p \\in P

       X_p \\geq 0\\text{, int} \\quad \\forall p \\in P

    The rental model minimizes cost of plans purchased
    with covering constraints:

    .. math::

       \\text{Min}  \\sum_{a \\in A} C_a
           \\sum_{p \\in P \\, \\mid \\, (p,a) \\in J} X_{(p,a)}

       \\text{s.t.} \\sum_{j \\in J \\, \\mid \\, f} X_j \\geq R_p
           \\quad \\forall p \\in P

       X_j \\geq 0\\text{, int} \\quad \\forall j \\in J

    The aggregate planning model minimizes production & holding costs
    while meeting demand over a number of periods:

    .. math::

       \\text{Min}  \\sum_{p \\in P} C_pX_p &+ hY_p

       \\text{s.t.} \\enspace Y_{p-1} + X_p - Y_p &= D_p
       &\\forall p \\in P

       Y_p &\\leq m &\\forall p \\in P

       Y_{\\min(P)-1} &= I_I

       Y_{\\max(P)} &= I_F

       X_p, \\, Y_p &\\geq 0\\text{, int} &\\forall p \\in P
    """
    if prob_class.lower() == "rental":
        return _rental(**kwargs)
    elif prob_class.lower() == "employee":
        return _employee(**kwargs)
    elif prob_class.lower() == "agg_planning":
        return _aggregate_planning(**kwargs)
    else:
        raise TypeError((
            "Invalid argument value {prob_class}: "
            "must be 'rental', 'employee', or 'job shop'.\n"))


def _rental(linear=False, **kwargs):
    """
    Factory method for the Rental scheduling problem.

    Parameters
    ----------
    linear : :py:obj:`bool`, optional
        Determines whether decision variables will be
        Reals (True) or Integer (False).
    **kwargs : optional
        if any given, returns pyomo concrete model instead, with these passed
        into pyomo's `create_instance`.
    """
    def _obj_expression(model):
        """Objective Expression: Minimizing Number of Workers"""
        my_expr = 0
        for (period, plan) in model.PlanToPeriod:
            my_expr += model.PlanCosts[plan] * model.NumRent[(period, plan)]
        return my_expr

    def _period_reqs_constraint_rule(model, p):
        """Constraints for having enough workers per period"""
        num_periods = len(model.Periods)
        my_sum = 0
        sum_terms = []
        for (period, plan) in model.PlanToPeriod:
            # Get index of current period
            ind = model.Periods.ord(p)
            # Get effective periods based on PlanLength
            periods_in_plan = [
                model.Periods[(
                    (ind - 1 - pl) % num_periods) + 1]
                for pl in range(model.PlanLengths[plan])]
            periods_in_plan = [per for per in periods_in_plan
                               if (per, plan) in model.PlanToPeriod]
            # Sum up how many rented in effective periods
            # new_terms makes sure that not adding same term again
            new_terms = [(p_in_plan, plan)
                         for p_in_plan in periods_in_plan
                         if (p_in_plan, plan) not in sum_terms]
            my_sum += sum([model.NumRent[term] for term in new_terms])
            sum_terms.extend(new_terms)
        return my_sum >= model.PeriodReqs[p]

    # Create the abstract model & dual suffix
    model = pyo.AbstractModel()
    model.dual = pyo.Suffix(direction=pyo.Suffix.IMPORT)
    # Define sets/params that are always used
    model.Periods = pyo.Set(ordered=True)
    model.Plans = pyo.Set()
    model.PlanToPeriod = pyo.Set(dimen=2)
    model.PeriodReqs = pyo.Param(model.Periods)
    model.PlanCosts = pyo.Param(model.Plans)
    model.PlanLengths = pyo.Param(model.Plans)
    # Define decision variables
    model.NumRent = pyo.Var(
        model.PlanToPeriod,
        within=pyo.NonNegativeReals if linear else pyo.NonNegativeIntegers)
    # Define objective & constraints
    model.OBJ = pyo.Objective(rule=_obj_expression, sense=pyo.minimize)
    model.PeriodReqsConstraint = pyo.Constraint(
        model.Periods,
        rule=_period_reqs_constraint_rule)
    # Check if returning concrete or abstract model
    if kwargs:
        return model.create_instance(**kwargs)
    else:
        return model


def _employee(linear=False, **kwargs):
    """
    Factory method for the Employee scheduling problem.

    Parameters
    ----------
    linear : :py:obj:`bool`, optional
        Determines whether decision variables will be
        Reals (True) or Integer (False).
    **kwargs : optional
        if any given, returns pyomo concrete model instead, with these passed
        into pyomo's `create_instance`.

    Notes
    -----
    Simple model: minimize # of workers employed to meet shift requirements
    """
    def _obj_expression(model):
        """Objective Expression: Minimizing Number of Workers"""
        return pyo.summation(model.NumWorkers)

    def _period_reqs_constraint_rule(model, p):
        """Constraints for having enough workers per period"""
        # Get index of current period
        ind = model.Periods.ord(p)
        num_periods = len(model.Periods)
        # Get effective periods based on ShiftLength - loops back
        effective_periods = [
            model.Periods[((ind - 1 - shift) % num_periods) + 1]
            for shift in range(model.ShiftLength.value)]
        # Sum up how many workers are working this period
        my_sum = sum([model.NumWorkers[period]
                      for period in effective_periods])
        return my_sum >= model.PeriodReqs[p]

    # Create the abstract model & dual suffix
    model = pyo.AbstractModel()
    model.dual = pyo.Suffix(direction=pyo.Suffix.IMPORT)
    # Define sets/params that are always used
    model.Periods = pyo.Set(ordered=True)
    model.ShiftLength = pyo.Param()  # num periods a worker works in a row
    model.PeriodReqs = pyo.Param(model.Periods)
    # Define decision variables
    model.NumWorkers = pyo.Var(
        model.Periods,
        within=pyo.NonNegativeReals if linear else pyo.NonNegativeIntegers)
    # Define objective & constraints
    model.OBJ = pyo.Objective(rule=_obj_expression, sense=pyo.minimize)
    model.PeriodReqsConstraint = pyo.Constraint(
        model.Periods,
        rule=_period_reqs_constraint_rule)
    # Check if returning concrete or abstract model
    if kwargs:
        return model.create_instance(**kwargs)
    else:
        return model


def _aggregate_planning(linear=False, **kwargs):
    """
    Factory method returning Pyomo Abstract/Concrete Model
    for the Aggregate Planning Problem

    Parameters
    ----------
    **kwargs
        Passed into Pyomo Abstract Model's `create_instance`
        to return Pyomo Concrete Model instead.

    Notes
    -----
    """
    def _obj_expression(model):
        """Objective Expression: """
        return (pyo.summation(model.Cost, model.Produce)
                + model.HoldingCost * pyo.summation(model.InvLevel))

    def _conserve_flow_constraint_rule(model, p):
        """Constraints for """
        ind = model.Periods.ord(p)
        if ind == 1:
            return (model.InitialInv + model.Produce[p]
                    - model.InvLevel[p] == model.Demand[p])
        else:
            last_p = model.Periods[ind - 1]
            return (model.InvLevel[last_p] + model.Produce[p]
                    - model.InvLevel[p] == model.Demand[p])

    def _max_storage_constraint_rule(model, p):
        return (0, model.InvLevel[p], model.MaxStorage)

    def _final_inv_constraint_rule(model):
        last_period = model.Periods[-1]
        return model.InvLevel[last_period] == model.FinalInv

    # Create the abstract model & dual suffix
    model = pyo.AbstractModel()
    model.dual = pyo.Suffix(direction=pyo.Suffix.IMPORT)
    # Define sets/params that are always used
    model.Periods = pyo.Set(ordered=True)
    model.Cost = pyo.Param(model.Periods)
    model.Demand = pyo.Param(model.Periods)
    model.HoldingCost = pyo.Param()
    model.MaxStorage = pyo.Param(within=pyo.Any, default=None)
    model.InitialInv = pyo.Param(default=0)
    model.FinalInv = pyo.Param(default=0)
    # Define decision variables
    model.Produce = pyo.Var(
        model.Periods,
        within=pyo.NonNegativeReals if linear else pyo.NonNegativeIntegers)
    model.InvLevel = pyo.Var(
        model.Periods,
        within=pyo.NonNegativeReals if linear else pyo.NonNegativeIntegers)
    # Define objective & constraints
    model.OBJ = pyo.Objective(rule=_obj_expression, sense=pyo.minimize)
    model.ConserveFlowConstraint = pyo.Constraint(
        model.Periods,
        rule=_conserve_flow_constraint_rule)
    model.MaxStorageConstraint = pyo.Constraint(
        model.Periods,
        rule=_max_storage_constraint_rule)
    model.FinalInvConstraint = pyo.Constraint(
        rule=_final_inv_constraint_rule)
    # Check if returning concrete or abstract model
    if kwargs:
        return model.create_instance(**kwargs)
    else:
        return model



"""
Module contains factory methods & other functions
for Mixed Integer Linear Programs (MILP).
"""
import math

import pyomo.environ as pyo
import pyomo
import pandas as pd


def blending(linear=True, **kwargs):
    """
    Factory method returning Pyomo Abstract/Concrete Model
    for the Blending Problem

    Parameters
    ----------
    linear : :py:obj:`bool`, optional
        Determines whether decision variables will be
        Reals (True) or Integer (False).
    **kwargs
        Passed into Pyomo Abstract Model's `create_instance`
        to return Pyomo Concrete Model instead.

    Notes
    -----
    The Blending Problem optimizes the mixing of ingredients
    to satisfy restrictions while minimizing cost.

    .. math::

       \\text{Min}  \\sum_{i \\in I} C_iX_i

       \\text{s.t.} \\enspace L_p \\leq \\sum_{i \\in I} N_{i,p}X_i
            \\leq U_p \\quad \\forall p \\in P

       \\sum_{i \\in I} X_i = 1

       X_i \\geq 0 \\quad \\forall i \\in I

    Examples
    --------
    Creating a concrete model from data params & solving instance.

    >>> import pyomo.environ as pyo
    >>> instance = blending("my_params.dat") # AMPL data file
    >>> opt = pyo.SolverFactory("glpk")
    >>> results = opt.solve(instance)
    """
    def _obj_expression(model):
        """Objective Expression: Maximizing Value"""
        return pyo.summation(model.Cost, model.Blend)

    def _property_constraint_rule(model, p):
        """Constraints for Scarce Resources"""
        return (model.MinProperty[p], sum(
            model.IngredientProperties[i, p]
            * model.Blend[i]
            for i in model.Ingredients
            ), model.MaxProperty[p])

    def _conservation_constraint_rule(model):
        return pyo.summation(model.Blend) == 1

    # Create the abstract model & dual suffix
    model = pyo.AbstractModel()
    model.dual = pyo.Suffix(direction=pyo.Suffix.IMPORT)
    # Define sets/params that are always used
    model.Ingredients = pyo.Set()
    model.Properties = pyo.Set()
    model.Cost = pyo.Param(model.Ingredients)
    model.IngredientProperties = pyo.Param(model.Ingredients, model.Properties)
    model.MinProperty = pyo.Param(model.Properties,
                                  within=pyo.Any, default=0.0)
    model.MaxProperty = pyo.Param(model.Properties,
                                  within=pyo.Any, default=None)
    # Define decision variables
    model.Blend = pyo.Var(
        model.Ingredients,
        within=pyo.NonNegativeReals if linear else pyo.NonNegativeIntegers,
        bounds=(0, None))
    # Define objective & constraints
    model.OBJ = pyo.Objective(rule=_obj_expression, sense=pyo.minimize)
    model.PropertyConstraint = pyo.Constraint(
        model.Properties,
        rule=_property_constraint_rule)
    model.ConservationConstraint = pyo.Constraint(
        rule=_conservation_constraint_rule)
    # Check if returning concrete or abstract model
    if kwargs:
        return model.create_instance(**kwargs)
    else:
        return model


def resource_allocation(
        linear=True, mult_res=False,
        max_activity=True, **kwargs):
    """
    Factory method returning Pyomo Abstract/Concrete Model
    for the Resource Allocation Problem.

    Parameters
    ----------
    linear : :py:obj:`bool`, optional
        Determines whether decision variables will be
        Reals (True) or Integer (False).
    mult_res : :py:obj:`bool`, optional
        Determines whether there are multiple of each resource or not.
    max_activity : :py:obj:`bool`, optional
        Determines whether there is an upper limit on the
        decision variables.
    **kwargs
        Passed into Pyomo Abstract Model's `create_instance`
        to return Pyomo Concrete Model instead.

    Notes
    -----
    The Resource Allocation Problem optimizes
    using scarce resources for valued activities.

    .. math::

       \\text{Max}  \\sum_{a \\in A} V_aX_a

       \\text{s.t.} \\sum_{a \\in A} N_{r,a}X_a \\leq M_r
           \\quad \\forall r \\in R

       0 \\leq X_a \\leq M_a \\quad \\forall a \\in A

    Examples
    --------
    Creating a concrete model from data params & solving instance.

    >>> import pyomo.environ as pyo
    >>> instance = resource_allocation("my_params.dat) # AMPL data file
    >>> opt = pyo.SolverFactory("glpk")
    >>> results = opt.solve(instance)
    """
    def _get_bounds(model, a):
        """Upper Bounds for the Decision Vars based on Maximum Demand."""
        return (0, model.MaxActivity[a])

    def _obj_expression(model):
        """Objective Expression: Maximizing Value"""
        return pyo.summation(model.Values, model.NumActivity)

    def _resource_constraint_rule(model, r):
        """Constraints for Scarce Resources"""
        return sum(
            model.ResourceNeeds[r, a]
            * model.NumActivity[a]
            for a in model.Activities
            ) <= model.MaxResource[r]

    def _mult_resource_constraint_rule(model, r):
        """Constraints for Scarce Resources"""
        return sum(
            model.ResourceNeeds[r, a]
            * model.NumActivity[a]
            for a in model.Activities
            ) <= model.MaxResource[r] * model.NumResource[r]
    # Create the abstract model & dual suffix
    model = pyo.AbstractModel()
    model.dual = pyo.Suffix(direction=pyo.Suffix.IMPORT)
    # Define sets/params that are always used
    model.Activities = pyo.Set()
    model.Resources = pyo.Set()
    model.Values = pyo.Param(model.Activities)
    model.ResourceNeeds = pyo.Param(model.Resources, model.Activities)
    model.MaxResource = pyo.Param(model.Resources)
    # If multiple resources, load those params
    if mult_res:
        model.NumResource = pyo.Param(model.Resources)
    # If upper bound on dec. vars, load those params
    if max_activity:
        model.MaxActivity = pyo.Param(model.Activities)
    # Define decision variables
    model.NumActivity = pyo.Var(
        model.Activities,
        within=pyo.NonNegativeReals if linear else pyo.NonNegativeIntegers,
        bounds=_get_bounds if max_activity else (0, None))
    # Define objective & resource constraints
    model.OBJ = pyo.Objective(rule=_obj_expression, sense=pyo.maximize)
    model.ResourceConstraint = pyo.Constraint(
        model.Resources,
        rule=_resource_constraint_rule if not mult_res
        else _mult_resource_constraint_rule)
    # Check if returning concrete or abstract model
    if kwargs:
        return model.create_instance(**kwargs)
    else:
        return model


def print_sol(instance, money_obj=False):
    """
    Print the solution to the solved `instance`.

    Parameters
    ----------
    instance : :py:class:`pyomo.environ.ConcreteModel`
        A solved model to retrieve objective & variable values.
    money_obj : bool
        Whether or not the objective is a monetary value (adds $ if True).

    Notes
    ----
    Assumes the objective is retrievable by :py:obj:`instance.OBJ()`
    """
    dollar_sign = "$" if money_obj else ""
    print(f"Objective Value: {dollar_sign}{instance.OBJ():,}")
    for v in instance.component_objects(pyo.Var, active=True):
        print("Variable component: ", v)
        for index in v:
            print("   ", index, v[index].value)


def sensitivity_analysis(instance):
    """
    Return dataframe containing sensitivity analysis

    Parameters
    ----------
    instance : :py:class:`pyomo.environ.ConcreteModel`
        A solved model to retrieve dual suffix.

    Returns
    -------
    pandas.DataFrame
        Dataframe with sensitivity analysis information.

    Notes
    -----
    Assumes the dual suffix is retrievable by :py:obj:`instance.dual`.
    """
    # Dual Variable Values
    dual_dict = {str(k): [v] for (k, v) in dict(instance.dual).items()}
    # Constraint Info
    for con in instance.dual:
        lower = con.lower if type(con.lower) \
            is not pyomo.core.expr.numvalue.NumericConstant \
            else con.lower.value
        upper = con.upper if type(con.upper) \
            is not pyomo.core.expr.numvalue.NumericConstant \
            else con.upper.value
        slack = 0 if math.isclose(con.slack(), 0,
                                  abs_tol=1e-5) else con.slack()
        active = True if slack == 0 else False
        vals = [lower, upper, slack, active]
        # Add constraint info to dual_dict
        dual_dict[con.name].extend(vals)
    # dual_dict to pandas dataframe
    sens_analysis = pd.DataFrame.from_dict(dual_dict,
                                           orient="index",
                                           columns=[
                                               "Dual Value",
                                               "Lower",
                                               "Upper",
                                               "Slack",
                                               "Active"])
    # Name the index and return the dataframe
    sens_analysis.index.name = "Constraint"
    return sens_analysis



from quantecon import MarkovChain
import numpy as np


def analyze_ctmc(states, rate_matrix, t=None, d=None, n=None, init=None):
    """
    Perform Markov Analysis of continuous time discrete state markov chain
    (CTMC) process.

    Parameters
    ----------
    states : array-like
        Vector-like of length M containing the values associated with the
        states, which must be homogeneous in type. If None, the values
        default to integers 0 through M-1.
    rate_matrix : array-like
        matrix of size MxM detailing the stationary probabilities of moving
        from one state to another.
    t : int
        Integer for the end time period for the transient probability analysis.
        If this is given, then either d or n must be given, and init must be
        given as well.
    d : float
        Float for the small delta (amount of time per step) for numerically
        solving the transient probabilities.  Either this or n (number of
        steps) must be given (one can be inferred from the other and 't').
    n : int
        Integer of the number of steps to take for numerically solving the
        transient probabilities.  Either this or 'd' must be given.
        If both are given and they do not coincide with t (d*n = t), an error
        will be raised.
    init : array-like
        Vector-like of length n containing the initial state values for
        the transient probability analyis.  This must be given if 't', 'd', or
        'n' is given.

    Returns
    -------
    analysis : dict
        Dictionary with results of markov analysis

    Raises
    ------
    ValueError
        If some but not all of the required transient probability analysis
        arguments are given, or if t, d, and n are given, but their values are
        invalid (t = n * d).
    """
    num_states = len(states)
    if init is not None:
        # transient solutions
        # have to use numerical approaches, no closed form in general
        # DTMC approximation (not embedded here)
        if (d is not None) and (n is None) and (t is not None):
            n = int(t / d)
        elif (d is None) and (n is not None) and (t is not None):
            d = t / n
        elif (d is not None) and (n is not None) and (t is None):
            t = n * d
        elif (d is not None) and (n is not None) and (t is not None):
            true_t = n * d
            if round(true_t) != t:
                raise ValueError("Given values for arguments 't', 'd', and"
                                 " 'n' are not compatible.  t must be equal"
                                 " to n * d.")
        else:
            raise ValueError("Transient analysis requires two of the three "
                             "arguments: 't' (end time), 'd' (delta), "
                             "'n' (num steps).")
        if init is None:
            raise ValueError("Argument 't' was provided without 'init'."
                             " Transient analysis requires 'init' for"
                             " the initial state vector.")
    if ((d is not None) or (n is not None) or
            (t is not None)) and (init is None):
        raise ValueError("Argument 'init' was not provided, but other"
                         " Transient analysis arguments were."
                         " 'init' is required for Transient analysis.")

    # transition_rates_i is sum of transition rates out of state i
    transition_rates = rate_matrix.sum(axis=1)  # sum across cols
    if t is not None:
        # P is state-transition matrix determined from rate matrix & d
        P = np.array([[1 - d * transition_rates[col] if row == col
                     else d * rate_matrix[row, col]
                     for col in states] for row in states])
        # prob that system in state i at time t: q_i(t)
        # q(t) = [q_0(t), q_1(t), ..., q_(m-1)(t)]
        # Note sum(q_t) == 1
        # transient sol. approx. at time t = n*d with DTMC
        #   by solving following equation:
        # eq = q(n*d) == q(0)P^(n)
        # or q(n*d + d) == q(n*d)P
        q_step = np.matmul(init, np.linalg.matrix_power(P, n))

    # Steady State probabilities
    # limit of q(t) as t goes to infinity
    # generator matrix
    gen_matrix = np.array([[-transition_rates[row] if row == col
                          else rate_matrix[row, col]
                          for col in states] for row in states])
    # augmented generator matrix - replace first col with 1s
    gen_matrix[:, 0] = np.ones(num_states)
    # Solve linear equations for steady state probs
    unit_vector = np.zeros(num_states)
    unit_vector[0] = 1
    steady_state = np.matmul(unit_vector.T, np.linalg.inv(gen_matrix))
    analysis = {'transition_rates': transition_rates, 'P': P, 'init': init,
                'transient': q_step,
                'generator_matrix': gen_matrix,
                'steady_state': steady_state}
    return analysis


def analyze_dtmc(P, states=None, sim_kwargs=None,
                 trans_kwargs=None, cost_kwargs=None):
    """
    Perform Markov Analysis of discrete time discrete state markov chain
    (DTMC) process.

    Parameters
    ----------
    P : array-like
        The transition matrix.  Must be of shape n x n.
    states : array-like
        Array_like of length n containing the values associated with the
        states, which must be homogeneous in type. If None, the values
        default to integers 0 through n-1.
    sim_kwargs : dict
        Dictionary of key word arguments to be passed to the simulation
        of the markov process.  If None, then no simulation will be performed.
        These include ts_length (length of each
        simulation) and init (Initial state values).
    trans_kwargs : dict
        Dictionary of options for the transient probability analysis (tsa).
        If None is passed instead of dict, no tsa will be done.
        ts_length is the number of time periods to analyze, while init is
        the initial state probability vector.
    cost_kwargs : dict
        Dictionary of cost parameters for cost analysis.  If None, then
        no cost analysis will be performed.  These include state (vector of
        costs of being in each state), transition (matrix of costs of
        transitioning from one state to another), and num (number of
        these processes - total cost multiplied by this, default 1).

    Returns
    -------
    analysis : dict
        Dictionary with results of markov analysis

    Raises
    ------
    ValueError
        If sim_kwargs, trans_kwargs, or cost_kwargs is given, but their
        required arguments are not passed.  These are described in the
        Notes section.

    Notes
    -----
    The required arguments if the kwargs are passed are:

    * sim_kwargs: `ts_length` is required, the length of the sim.

    * trans_kwargs: `ts_length` is required, the number of periods
      to analyze

    * cost_kwargs: `state` and `transition` are required, which are the
      costs of being in any state and the costs of transitioning from one
      state to another.
    """
    analysis = {}
    markov = MarkovChain(P, states)
    analysis["cdfs"] = markov.cdfs
    steady_state = markov.stationary_distributions[0]
    analysis["steady_state"] = {"output": steady_state}
    if sim_kwargs:
        if "ts_length" not in sim_kwargs:
            raise ValueError(("Required argument `ts_length` in sim_kwargs! "
                              "None was given."))
        if "init" not in sim_kwargs:
            sim_kwargs["init"] = None
        analysis["sim"] = {"kwargs": sim_kwargs,
                           "output": markov.simulate(**sim_kwargs)}
    if trans_kwargs:
        if "ts_length" not in trans_kwargs:
            raise ValueError(("Required argument `ts_length` in trans_kwargs! "
                              "None was given."))
        if "init" not in trans_kwargs:
            trans_kwargs["init"] = None
        trans_probs = _transient_probs(P, states, **trans_kwargs)
        analysis["transient"] = {"kwargs": trans_kwargs,
                                 "output": trans_probs}
    if cost_kwargs:
        if "state" not in cost_kwargs:
            raise ValueError(("Required argument `state` in trans_kwargs! "
                              "None was given."))
        if "transition" not in cost_kwargs:
            raise ValueError(("Required argument `transition` in trans_kwargs!"
                              " None was given."))
        if "num" not in cost_kwargs:
            cost_kwargs["num"] = 1
        # Cost of steady state
        cost_vector, cost_total = _cost_analysis(P, steady_state,
                                                 **cost_kwargs)
        # Cost of transient analysis
        analysis["steady_state"]["cost"] = \
            {"kwargs": cost_kwargs,
             "total": cost_total, "vector": cost_vector}
        if trans_kwargs:
            cost_vector, cost_total = _cost_analysis(P, trans_probs,
                                                     **cost_kwargs)
            analysis["transient"]["cost"] = {"kwargs": cost_kwargs,
                                             "total": cost_total,
                                             "vector": cost_vector}
    return analysis


def _cost_analysis(P, probs, state, transition, num):
    """
    Cost analysis for markov process

    probs could be transient or steady state
    """
    # Get cost vector - element wise mult, sum across cols
    cost_vector = state + np.sum(transition * P, 1)

    # Calculate expected costs
    exp_cost = np.matmul(probs, cost_vector)
    total_cost = np.sum(exp_cost) * num
    return exp_cost, total_cost


def _transient_probs(P, states, ts_length, init=None):
    """
    Calculate transient probabilities

    q(n): probability dist at time n
        q(n) = q(n-1) * P
    """
    if init is None:
        init = [np.random.choice(states, size=P.shape[0])]
    q = np.array([init])
    # q_n = q_(n-1) * P
    for _ in range(1, ts_length + 1):
        q = np.vstack((q, np.matmul(q[-1], P)))
    return q


def print_markov(analysis, mtype="dtmc"):
    """
    Print results of markov analysis.

    Parameters
    ----------
    analysis : dict
        dictionary returned from markov_analysis() containing
        cdfs, steady state probs, etc.
    mtype : str
        Type of markov analysis that is being passed.  Must be
        'dtmc' or 'ctmc'.

    Raises
    ------
    ValueError
        If invalid value is passed for 'mtype'.
    """
    if mtype == "dtmc":
        print("CDFs:")
        print(analysis["cdfs"])
        print()
        print("Steady State Probs:")
        print(analysis["steady_state"]['output'])
        print()
        if "sim" in analysis:
            print(("Simulation of length "
                   f"{analysis['sim']['kwargs']['ts_length']}"))
            print("Initial Conditions:")
            print(analysis["sim"]["kwargs"]["init"])
            print("Output:")
            print(analysis["sim"]["output"])
            print()
        if "transient" in analysis:
            print(("Transient Probabilities (length "
                  f"{analysis['transient']['kwargs']['ts_length']})"))
            print("Initial Conditions:")
            print(analysis["transient"]["kwargs"]["init"])
            print("Output:")
            print(analysis["transient"]["output"])
            print()
        if "cost" in analysis['steady_state']:
            print("Cost kwargs:")
            print(analysis['steady_state']['cost']['kwargs'])
            print("Expected Steady State Cost:")
            print(f"${analysis['steady_state']['cost']['vector']:,.2f}")
            print(("Expected Total Steady State Cost: $"
                  f"{analysis['steady_state']['cost']['total']:,.2f}"))
        if "transient" in analysis:
            if "cost" in analysis['transient']:
                print("Expected Transient Cost:")
                print(analysis["transient"]['cost']['vector'])
                print(("Expected Total Transient Cost: $"
                      f"{analysis['transient']['cost']['total']:,.2f}"))
    elif mtype == "ctmc":
        print("Transition Rates:")
        print(analysis["transition_rates"])
        print("P:")
        print(analysis["P"])
        print("Steady State:")
        print(analysis["steady_state"])
        if "transient" in analysis:
            print("Initial States:")
            print(analysis["init"])
            print("Transient Probabilities:")
            print(analysis["transient"])
            print("Generator Matrix:")
            print(analysis["generator_matrix"])
    else:
        raise ValueError(f"Invalid value for mtype: {mtype}.  Must be "
                         "'dtmc' or 'ctmc'.")



import io
import sys

import pyomo.environ as pyo
import pandas as pd

from ormm.mathprog import resource_allocation, \
    print_sol, sensitivity_analysis, blending
from tests.methods import solve_instance

MATHPROG_DATA = "ormm/mathprog/example_data/"
SIMPLE_RES_DATA = MATHPROG_DATA + "resource_allocation.dat"
COMPLEX_RES_DATA = MATHPROG_DATA + "mult_resource_allocation.dat"
BLENDING_DATA = MATHPROG_DATA + "blending.dat"


def test_simple_resource_allocation_with_data():
    model = resource_allocation()
    instance1 = model.create_instance(SIMPLE_RES_DATA)
    instance2 = resource_allocation(filename=SIMPLE_RES_DATA)
    for inst in [instance1, instance2]:
        instance, results = solve_instance(inst)
        # Check all variable values
        for v in instance.component_objects(pyo.Var, active=True):
            if v.name == "NumActivity":
                assert {index: v[index].value
                        for index in v} == {"Q": 31.25, "W": 75}
        assert instance.OBJ() == 7343.75


def test_complex_resource_allocation_with_data():
    instance = resource_allocation(
        filename=COMPLEX_RES_DATA, mult_res=True, max_activity=False)
    instance, results = solve_instance(instance)
    # Check all variable values
    for v in instance.component_objects(pyo.Var, active=True):
        if v.name == "NumActivity":
            assert {
                index: round(v[index].value, 2) for index in v} == {
                "Q": 123.08, "W": 0, "E": 0, "R": 0, "T": 46.15, "Y": 0}
    assert round(instance.OBJ(), 0) == 2692


def test_blending():
    model = blending()
    instance1 = model.create_instance(BLENDING_DATA)
    instance2 = blending(filename=BLENDING_DATA)
    for inst in [instance1, instance2]:
        instance, results = solve_instance(inst)
        for v in instance.component_objects(pyo.Var, active=True):
            if v.name == "Blend":
                assert {
                    index: round(v[index].value, 2) for index in v} == {
                    "Banana": 0.02, "Milk": 0.76, "Yogurt": 0.22}
        assert round(instance.OBJ(), 2) == 37.61


def test_print_sol_with_data():
    instance = resource_allocation(filename=SIMPLE_RES_DATA)
    instance, results = solve_instance(instance)
    # Redirect output to StringIO object
    captured_output = io.StringIO()
    sys.stdout = captured_output
    print_sol(instance, money_obj=True)
    sys.stdout = sys.__stdout__  # reset stdout
    test_string = (
        "Objective Value: $7,343.75\n"
        "Variable component:  NumActivity\n"
        "    Q 31.25\n"
        "    W 75.0\n"
    )
    assert captured_output.getvalue() == test_string


def test_sensitivity_analysis():
    instance = resource_allocation(
        filename=COMPLEX_RES_DATA, mult_res=True, max_activity=False)
    instance, results = solve_instance(instance)
    sens_analysis_df = sensitivity_analysis(instance)
    test_df = pd.DataFrame({
        "Dual Value": [9.29, 0.00, 5.22, 0.00],
        "Lower": [None, None, None, None],
        "Upper": [200.0, 280.0, 160.0, 320.0],
        "Slack": [0, 113.85, 0, 107.69],
        "Active": [True, False, True, False]
        }, index=[
            "ResourceConstraint[A]",
            "ResourceConstraint[B]",
            "ResourceConstraint[C]",
            "ResourceConstraint[D]"])
    sens_analysis_df["Dual Value"] = \
        sens_analysis_df["Dual Value"].round(decimals=2)
    sens_analysis_df["Slack"] = \
        sens_analysis_df["Slack"].round(decimals=2)
    assert sens_analysis_df.equals(test_df)



import time
import random

import pyomo.environ as pyo
import pandas as pd

from ormm.network import transportation_model, Graph
from tests.methods import solve_instance

# Folders and Files
NETWORK_DATA = "ormm/network/example_data/"
TRANSPORTATION_DATA = NETWORK_DATA + "transportation.dat"

# Data for Shortest Path Problem
SIMPLE_ARCS = [["A", "B", 7, "one"],
               ["B", "C", 3, "one"],
               ["A", "D", 5, "one"],
               ["D", "E", 5, "one"],
               ["E", "F", 5, "one"],
               ["F", "C", 10, "two"]]
SIMPLE_ARCS_NO_DIRECTION = [arc[:-1] for arc in SIMPLE_ARCS]

# Data for Transportation Model
SUPPLY = {"S1": 15, "S2": 15, "S3": 15}
DEMAND = {"D1": 5, "D2": 10, "D3": 15, "D4": 5, "D5": 10}
ARC_DATA = pd.DataFrame([["S1", 15, 15, 16, 11, 11],
                         ["S2", 13, 11, 15,  9,  6],
                         ["S3", 8,  12, 11,  7,  8]],
                        columns=["From", "D1", "D2", "D3", "D4", "D5"])
ARC_DATA = pd.melt(ARC_DATA, id_vars=["From"],
                   value_vars=["D1", "D2", "D3", "D4", "D5"],
                   var_name="To", value_name="Cost")
ARC_DATA["Direction"] = "one"


def test_transportation_model():
    model = transportation_model()
    instance1 = model.create_instance(TRANSPORTATION_DATA)
    instance2 = transportation_model(filename=TRANSPORTATION_DATA)
    graph = Graph()
    graph.add_arcs(ARC_DATA)
    instance_graph = graph.transportation(SUPPLY, DEMAND)
    for inst in [instance1, instance2, instance_graph]:
        instance, results = solve_instance(inst)
        assert instance.OBJ() == 475
        for v in instance.component_objects(pyo.Var, active=True):
            assert {index: v[index].value
                    for index in v} == {
                        ('S1', 'D1'): 0.0, ('S1', 'D2'): 5.0,
                        ('S1', 'D3'): 5.0, ('S1', 'D4'): 5.0,
                        ('S1', 'D5'): 0.0, ('S2', 'D1'): 0.0,
                        ('S2', 'D2'): 5.0, ('S2', 'D3'): 0.0,
                        ('S2', 'D4'): 0.0, ('S2', 'D5'): 10.0,
                        ('S3', 'D1'): 5.0, ('S3', 'D2'): 0.0,
                        ('S3', 'D3'): 10.0, ('S3', 'D4'): 0.0,
                        ('S3', 'D5'): 0.0}


def huge_transportation_model():
    # Define large random datasets
    num_nodes = 1_000
    upper_lim = 100
    source_nodes = {None: [f"S{x}" for x in range(num_nodes)]}
    dest_nodes = {None: [f"D{x}" for x in range(num_nodes)]}
    supply = {s: 5 for s in source_nodes[None]}
    demand = {d: 5 for d in dest_nodes[None]}
    ship_costs = {(s, d): random.randint(0, upper_lim)
                  for s in source_nodes[None] for d in dest_nodes[None]}
    my_data = {"Sources": source_nodes,
               "Destinations": dest_nodes,
               "Supply": supply,
               "Demand": demand,
               "ShippingCosts": ship_costs}
    # Define model, load data in it
    model = transportation_model()
    inst = model.create_instance(data={None: my_data})
    # Start timer and solve
    start_time = time.time()
    instance, results = solve_instance(inst)
    end_time = time.time()
    # Print termination status and solving time
    print(results.solver.termination_condition)
    print(end_time - start_time)


def test_graph_repr():
    graph = Graph()
    graph.add_arcs(SIMPLE_ARCS)
    test_str = ("Graph(defaultdict(<class 'list'>, {'A': ['B', 'D'], "
                "'B': ['C'], 'D': ['E'], 'E': ['F'], 'F': ['C'], "
                "'C': ['F']}), {('A', 'B'): 7, ('B', 'C'): 3, "
                "('A', 'D'): 5, ('D', 'E'): 5, ('E', 'F'): 5, "
                "('F', 'C'): 10, ('C', 'F'): 10}, "
                "['A', 'B', 'C', 'D', 'E', 'F'])")
    graph_repr = repr(graph)
    # Need to reorder set of nodes at end
    set_start_index = graph_repr.rfind("{")
    graph_repr_ordered = graph_repr[:set_start_index] + \
        str(sorted(eval(graph_repr[set_start_index:-1]))) + ")"
    assert graph_repr_ordered == test_str


def test_add_arcs_input_types():
    list_tuple = [tuple(arc) for arc in SIMPLE_ARCS]
    tuple_tuple = tuple(tuple(arc) for arc in SIMPLE_ARCS)
    df_no_cols_direction = pd.DataFrame(SIMPLE_ARCS)
    df_cols_direction = pd.DataFrame(
        SIMPLE_ARCS, columns=["From", "To", "Cost", "Direction"])
    df_no_cols_no_direction = pd.DataFrame(SIMPLE_ARCS_NO_DIRECTION)
    df_cols_no_direction = pd.DataFrame(
        SIMPLE_ARCS_NO_DIRECTION, columns=["From", "To", "Cost"])
    dict_arcs = {"From": [arc[0] for arc in SIMPLE_ARCS],
                 "To": [arc[1] for arc in SIMPLE_ARCS],
                 "Cost": [arc[2] for arc in SIMPLE_ARCS],
                 "Direction": [arc[3] for arc in SIMPLE_ARCS]}
    input_types = [list_tuple, tuple_tuple,
                   df_no_cols_direction, df_cols_direction,
                   df_no_cols_no_direction, df_cols_no_direction, dict_arcs]
    for input_type in input_types:
        graph = Graph()
        graph.add_arcs(input_type)
        analysis = graph.shortest_path("A")
        costs, paths = analysis["Costs"], analysis["Paths"]
        test_costs = {'A': 0, 'D': 5, 'B': 7, 'E': 10, 'C': 10, 'F': 15}
        test_paths = {'A': ('A',), 'D': ('A', 'D'), 'B': ('A', 'B'),
                      'E': ('A', 'D', 'E'), 'C': ('A', 'B', 'C'),
                      'F': ('A', 'D', 'E', 'F')}
        assert costs == test_costs
        assert paths == test_paths


def test_shortest_path_simple():
    graph = Graph()
    graph.add_arcs(SIMPLE_ARCS)
    analysis = graph.shortest_path("A")
    costs, paths = analysis["Costs"], analysis["Paths"]
    test_costs = {'A': 0, 'D': 5, 'B': 7, 'E': 10, 'C': 10, 'F': 15}
    test_paths = {'A': ('A',), 'D': ('A', 'D'), 'B': ('A', 'B'),
                  'E': ('A', 'D', 'E'), 'C': ('A', 'B', 'C'),
                  'F': ('A', 'D', 'E', 'F')}
    assert costs == test_costs
    assert paths == test_paths


if __name__ == "__main__":
    # huge_transportation_model()
    test_transportation_model()
    test_shortest_path_simple()



import io
import sys

from quantecon.markov import MarkovChain
import scipy.stats
import numpy as np
import pytest

from ormm.markov import analyze_dtmc, print_markov, analyze_ctmc


def test_income_audit():
    """A simple example of markov chain"""
    # Define MarkovChain object - P is transition matrix
    P = [[0.6, 0.4], [0.5, 0.5]]
    state_values = [0, 1]
    analysis = analyze_dtmc(P, state_values,
                            sim_kwargs={"ts_length": 25,
                                        "random_state": 42})
    analysis["steady_state"]['output'] = \
        analysis["steady_state"]['output'].round(3)
    test = {
        'cdfs': np.array([[0.6, 1.],
                          [0.5, 1.]]),
        'steady_state': {'output': np.array([[0.556, 0.444]])},
        'sim': {'kwargs': {'ts_length': 25, 'init': None, "random_state": 42},
                'output': np.array([0, 1, 0, 1, 1, 0, 0, 0, 0, 0, 1, 0, 1, 1,
                                    0, 1, 1, 1, 0, 0, 0, 0, 0, 1, 0])}}
    # Assert that numpy arrays are the same (except sim & stdy state output)
    assert all([np.allclose(analysis[k], test[k])
                for k in analysis if k not in ['sim', 'steady_state']])
    # assert sim kwargs values are the same
    assert all([test["sim"]["kwargs"][k] == analysis["sim"]["kwargs"][k]
                for k in analysis["sim"]["kwargs"]])
    # Assert that sim output numpy arrays are the same (fixed random state)
    assert np.allclose(analysis["sim"]["output"], test["sim"]["output"])
    # Assert that steady state output numpy arrays are the same
    assert np.allclose(analysis["steady_state"]["output"],
                       test["steady_state"]["output"])
    # Assert that both have the same keys
    assert test.keys() == analysis.keys()
    assert test['sim'].keys() == analysis["sim"].keys()
    assert test['sim']["kwargs"].keys() == analysis["sim"]["kwargs"].keys()

    # Assert print_markov works correctly with sim kwarg
    captured_output = io.StringIO()
    sys.stdout = captured_output
    print_markov(analysis)
    sys.stdout = sys.__stdout__  # reset stdout
    test_str = ("CDFs:\n"
                "[[0.6 1. ]\n"
                " [0.5 1. ]]\n\n"
                "Steady State Probs:\n"
                "[0.556 0.444]\n\n"
                "Simulation of length 25\n"
                "Initial Conditions:\n"
                "None\n"
                "Output:\n"
                "[0 1 0 1 1 0 0 0 0 0 1 0 1 1 0 1 1 1 0 0 0 0 0 1 0]\n\n")
    assert captured_output.getvalue() == test_str

    # Assert ValueErrors are raised
    with pytest.raises(ValueError):
        analysis = analyze_dtmc(P, state_values, sim_kwargs={"no_ts_len": 0})
    with pytest.raises(ValueError):
        analysis = analyze_dtmc(P, state_values, trans_kwargs={"no_ts_len": 0})
    with pytest.raises(ValueError):
        analysis = analyze_dtmc(P, state_values, cost_kwargs={"no_ts_len": 0})
    with pytest.raises(ValueError):
        analysis = analyze_dtmc(P, state_values,
                                cost_kwargs={"no_transition": 0})
    with pytest.raises(ValueError):
        analysis = analyze_dtmc(P, state_values,
                                cost_kwargs={"no_state": 0, "transition": 0})


def test_computer_repair():
    """
    One worker requires 2 days to repair a machine, and there are 2 of them.
    They fail as independent events: 0.2 to fail, 0.8 to not fail within
        the day.
    We use the binomial distribution to model the event of 0, 1, or 2 failures.
    There are 5 possible states, which we will define as (s_1, s_2):
        s_1 = num days first machine been in shop
        s_2 = num days second has been in shop
        s_1 = 0 if machine not failed, 1 if in first day of repair, 2 if second
        The 5 possible states: (0,0), (1,0), (2,0), (1,1), (2,1)
    """
    # Failure of computers is binomial
    #   p = 0.2 and n = 2
    pmf = scipy.stats.binom.pmf(k=[0, 1, 2], n=2, p=0.2)
    p_0, p_1, p_2 = pmf[0], pmf[1], pmf[2]
    # Define trainsition matrix
    transition_matrix = [[p_0, p_1, 0.0, p_2, 0.0],
                         [0.0, 0.0, 0.8, 0.0, 0.2],
                         [0.8, 0.2, 0.0, 0.0, 0.0],
                         [0.0, 0.0, 0.0, 0.0, 1.0],
                         [0.0, 1.0, 0.0, 0.0, 0.0]]
    states = [(0, 0), (1, 0), (2, 0), (1, 1), (2, 1)]
    analysis = analyze_dtmc(P=transition_matrix, states=states)
    captured_output = io.StringIO()
    sys.stdout = captured_output
    print_markov(analysis)
    sys.stdout = sys.__stdout__  # reset stdout
    test_str = ("CDFs:\n"
                "[[0.64 0.96 0.96 1.   1.  ]\n"
                " [0.   0.   0.8  0.8  1.  ]\n"
                " [0.8  1.   1.   1.   1.  ]\n"
                " [0.   0.   0.   0.   1.  ]\n"
                " [0.   1.   1.   1.   1.  ]]\n\n"
                "Steady State Probs:\n"
                "[0.45351474 0.25510204 0.20408163 0.01814059"
                " 0.069161  ]\n\n")
    assert captured_output.getvalue() == test_str


def test_light_bulb_replacement():
    """
    How many of the bulbs on average will have to be replaced each month?
        How much the budget for the repairs should be?
    Time of fauliure is uncertain
        variety of maintenance optioins
    failed bulbs are replaced monthly - use this for time intervals for dist
    """
    # Define state space, the age of the bulb
    state_space = [0, 1, 2, 3, 4]  # new, 1 month, 2 month, etc.
    num_states = len(state_space)
    num_bulbs = 1_000
    # probability of bulb failing based on age of bulb in months
    prob = [0.5, 0.1, 0.1, 0.1, 0.2]
    cdf = np.cumsum(prob)
    cond_prob = [p / (1 - cdf[ind - 1])
                 if ind > 0 else p for ind, p in enumerate(prob)]
    cond_prob[-1] = 1  # roundoff error overriding
    transition_matrix = np.zeros(shape=(num_states, num_states))
    transition_matrix[:, 0] = cond_prob
    for row in range(num_states - 1):
        transition_matrix[row, row + 1] = 1 - transition_matrix[row, 0]

    # Costs of inspecting & replacing light bulbs
    inspect_cost = 0.10
    replace_cost = 2
    test_cost_vector = [inspect_cost + replace_cost * transition_matrix[x, 0]
                        for x in range(num_states)]
    inspect_vector = [inspect_cost] * num_states
    replace_matrix = np.array([[replace_cost] * num_states]
                              + ([[0] * num_states] * (num_states - 1))).T

    # Transient probabilities
    #  all bulbs start at age 0 (new sign)
    #  estimate how many bulbs will be replaced during each of first 12 months?
    # q(n): probability dist for age of bulb at time n
    test_q = np.array([[1, 0, 0, 0, 0]])  # row 0 is q(0), row 1 is q(1), etc.
    # q_n = q_(n-1) * P
    for _ in range(1, 13):  # for the 12 months
        test_q = np.vstack((test_q, np.matmul(test_q[-1], transition_matrix)))
    # Calculate expected transient costs - don't include month 0
    exp_trans_cost = np.delete(np.matmul(test_q, test_cost_vector), 0)
    test_trans_cost = sum(exp_trans_cost) * num_bulbs
    print("test trans cost: ", test_trans_cost)
    num_replaced = int(sum(test_q[1:, 0]) * num_bulbs)
    print("Num Replaced: ", num_replaced)

    # Steady state probabilities
    # for discrete-time markov chains, as long as each state can be
    #  reached from every other state, the transient probabilities
    #  will approach equilibrium.
    markov_obj = MarkovChain(P=transition_matrix, state_values=state_space)
    test_steady_state = markov_obj.stationary_distributions
    print('steady state: ', test_steady_state)

    analysis = analyze_dtmc(transition_matrix, state_space,
                            trans_kwargs={"ts_length": 12,
                                          "init": [1, 0, 0, 0, 0]},
                            cost_kwargs={"state": inspect_vector,
                                         "transition": replace_matrix,
                                         "num": num_bulbs})
    test = {
        'cdfs': np.array(
            [[0.5, 1., 1., 1., 1.],
             [0.2, 0.2, 1., 1., 1.],
             [0.25, 0.25, 0.25, 1., 1.],
             [0.33333333, 0.33333333, 0.33333333, 0.33333333, 1.],
             [1., 1., 1., 1., 1.]]),
        'steady_state': {
            'cost': {'kwargs': {'num': 1000,
                                'state': [0.1, 0.1, 0.1, 0.1, 0.1],
                                'transition': np.array([[2, 0, 0, 0, 0],
                                                        [2, 0, 0, 0, 0],
                                                        [2, 0, 0, 0, 0],
                                                        [2, 0, 0, 0, 0],
                                                        [2, 0, 0, 0, 0]])},
                     'total': 933.3333333333333,
                     'vector': 0.9333333333333332},
            'output': np.array(
                [0.41666667, 0.20833333, 0.16666667, 0.125, 0.08333333])},
        'transient': {
            'cost': {'kwargs': {'num': 1000,
                                'state': [0.1, 0.1, 0.1, 0.1, 0.1],
                                'transition': np.array([[2, 0, 0, 0, 0],
                                                        [2, 0, 0, 0, 0],
                                                        [2, 0, 0, 0, 0],
                                                        [2, 0, 0, 0, 0],
                                                        [2, 0, 0, 0, 0]])},
                     'total': 12009.303421875004,
                     'vector': np.array([1.1, 0.8, 0.75, 0.795, 1.0825,
                                         0.99575, 0.920625, 0.8976375,
                                         0.90770625, 0.95175438,
                                         0.94762406, 0.93364684,
                                         0.92705939])},
            'kwargs': {'init': [1, 0, 0, 0, 0], 'ts_length': 12},
            'output': np.array(
                [[1., 0., 0., 0., 0.],
                 [0.5, 0.5, 0., 0., 0.],
                 [0.35, 0.25, 0.4, 0., 0.],
                 [0.325, 0.175, 0.2, 0.3, 0.],
                 [0.3475, 0.1625, 0.14, 0.15, 0.2],
                 [0.49125, 0.17375, 0.13, 0.105, 0.1],
                 [0.447875, 0.245625, 0.139, 0.0975, 0.07],
                 [0.4103125, 0.2239375, 0.1965, 0.10425, 0.065],
                 [0.39881875, 0.20515625, 0.17915, 0.147375, 0.0695],
                 [0.40385313, 0.19940938, 0.164125, 0.1343625, 0.09825],
                 [0.42587719, 0.20192656, 0.1595275, 0.12309375, 0.089575],
                 [0.42381203, 0.21293859, 0.16154125, 0.11964563, 0.0820625],
                 [0.41682342, 0.21190602, 0.17035088, 0.12115594, 0.07976375]]
                )}}
    # Assert two dictionaries have equal values
    assert all(is_analysis_equal(analysis, test))
    # Assert that both have the same keys
    assert test.keys() == analysis.keys()
    assert test['steady_state'].keys() == analysis["steady_state"].keys()
    assert test['steady_state']['cost'].keys() == \
        analysis['steady_state']['cost'].keys()
    assert test['steady_state']['cost']['kwargs'].keys() == \
        analysis['steady_state']['cost']['kwargs'].keys()

    assert test['transient'].keys() == analysis["transient"].keys()
    assert test['transient']["kwargs"].keys() == \
        analysis["transient"]["kwargs"].keys()
    assert test['transient']['cost'].keys() == \
        analysis['transient']['cost'].keys()
    assert test['transient']['cost']['kwargs'].keys() == \
        analysis['transient']['cost']['kwargs'].keys()

    # Test print_markov
    captured_output = io.StringIO()
    sys.stdout = captured_output
    print_markov(analysis)
    sys.stdout = sys.__stdout__  # reset stdout
    test_str = ("CDFs:\n"
                "[[0.5        1.         1.         1.         1.        ]\n"
                " [0.2        0.2        1.         1.         1.        ]\n"
                " [0.25       0.25       0.25       1.         1.        ]\n"
                " [0.33333333 0.33333333 0.33333333 0.33333333 1.        ]\n"
                " [1.         1.         1.         1.         1.        ]]\n"
                "\nSteady State Probs:\n"
                "[0.41666667 0.20833333 0.16666667 0.125      0.08333333]\n\n"
                "Transient Probabilities (length 12)\n"
                "Initial Conditions:\n"
                "[1, 0, 0, 0, 0]\n"
                "Output:\n"
                "[[1.         0.         0.         0.         0.        ]\n"
                " [0.5        0.5        0.         0.         0.        ]\n"
                " [0.35       0.25       0.4        0.         0.        ]\n"
                " [0.325      0.175      0.2        0.3        0.        ]\n"
                " [0.3475     0.1625     0.14       0.15       0.2       ]\n"
                " [0.49125    0.17375    0.13       0.105      0.1       ]\n"
                " [0.447875   0.245625   0.139      0.0975     0.07      ]\n"
                " [0.4103125  0.2239375  0.1965     0.10425    0.065     ]\n"
                " [0.39881875 0.20515625 0.17915    0.147375   0.0695    ]\n"
                " [0.40385313 0.19940938 0.164125   0.1343625  0.09825   ]\n"
                " [0.42587719 0.20192656 0.1595275  0.12309375 0.089575  ]\n"
                " [0.42381203 0.21293859 0.16154125 0.11964563 0.0820625 ]\n"
                " [0.41682342 0.21190602 0.17035088 0.12115594 0.07976375]]\n"
                "\nCost kwargs:\n"
                "{'state': [0.1, 0.1, 0.1, 0.1, 0.1], 'transition': "
                "array([[2, 0, 0, 0, 0],\n"
                "       [2, 0, 0, 0, 0],\n"
                "       [2, 0, 0, 0, 0],\n"
                "       [2, 0, 0, 0, 0],\n"
                "       [2, 0, 0, 0, 0]]), 'num': 1000}\n"
                "Expected Steady State Cost:\n"
                "$0.93\n"
                "Expected Total Steady State Cost: $933.33\n"
                "Expected Transient Cost:\n"
                "[1.1        0.8        0.75       0.795      1.0825"
                "     0.99575\n"
                " 0.920625   0.8976375  0.90770625 0.95175438 0.94762406"
                " 0.93364684\n"
                " 0.92705939]\n"
                "Expected Total Transient Cost: $12,009.30\n")
    assert captured_output.getvalue() == test_str


def is_analysis_equal(analysis, test):
    is_equal = []
    for k in analysis:
        if type(analysis[k]) is dict:
            is_equal.extend(is_analysis_equal(analysis[k], test[k]))
        elif type(analysis[k]) in [np.float64, float, int]:
            is_equal.append(round(analysis[k], 5) == round(test[k], 5))
        elif type(analysis[k]) is str:
            is_equal.append(analysis[k] == test[k])
        elif type(analysis[k]) is np.ndarray:
            is_equal.append(np.allclose(analysis[k], test[k]))
        elif type(analysis[k]) is list:
            is_equal.append(all(a == b for a, b in zip(analysis[k], test[k])))
    return is_equal


def test_atm_example():
    arrival_rate = 2  # per minute
    service_rate = 2.5  # per minute
    states = [0, 1, 2, 3, 4, 5]
    num_states = len(states)
    rate_matrix = []
    for row in states:
        this_row = [arrival_rate if col == row + 1 else 0 for col in states]
        if row >= 1:
            this_row[row - 1] = service_rate
        rate_matrix.append(this_row)
    rate_matrix = np.array(rate_matrix)

    # transient solutions
    # have to use numerical approcaches, no closed form in general
    # QuantEcon only has finite state discrete time MarkovChain
    # DTMC approximation (not embedded here)
    d = 0.05  # small time interval, a parameter
    t = 1  # want to approx. transient probs at this time, a parameter
    n = int(t / d)  # number of steps - determines accuracy of approx.
    # alpha_i is sum of transition rates out of state i
    alpha = rate_matrix.sum(axis=1)  # sum across cols
    # P is state-transition matrix determined from rate matrix & d
    P = [[1 - d*alpha[col] if row == col else d * rate_matrix[row, col]
          for col in states] for row in states]
    P = np.array(P)
    # prob that system in state i at time t: q_i(t)
    # q(t) = [q_0(t), q_1(t), ..., q_(m-1)(t)]
    # Note sum(q_t) == 1
    # Initial probability vector:
    q_init = [1, 0, 0, 0, 0, 0]
    # transient sol. approx. at time t = n*d with DTMC
    #   by solving following equation:
    # eq = q(n*d) == q(0)P^(n)
    # or q(n*d + d) == q(n*d)P
    q_step = np.matmul(q_init, np.linalg.matrix_power(P, n))

    # My analysis
    analysis = analyze_ctmc(states=states, rate_matrix=rate_matrix,
                            t=t, d=d, init=q_init)
    assert (analysis['transient'] == q_step).all()

    # Steady State probabilities
    # limit of q(t) as t goes to infinity
    # generator matrix
    gen_matrix = [[-alpha[row] if row == col else rate_matrix[row, col]
                   for col in states] for row in states]
    gen_matrix = np.array(gen_matrix)
    # augmented generator matrix - replace first col with 1s
    gen_matrix[:, 0] = np.ones(num_states)
    # Solve linear equations for steady state probs
    unit_vector = np.zeros(num_states)
    unit_vector[0] = 1
    steady_state = np.matmul(unit_vector.T, np.linalg.inv(gen_matrix))
    assert (analysis['steady_state'] == steady_state).all()

    no_d_analysis = analyze_ctmc(states=states, rate_matrix=rate_matrix,
                                 t=t, n=n, init=q_init)
    assert all(is_analysis_equal(analysis, no_d_analysis))
    no_t_analysis = analyze_ctmc(states=states, rate_matrix=rate_matrix,
                                 n=n, d=d, init=q_init)
    assert all(is_analysis_equal(analysis, no_t_analysis))

    with pytest.raises(ValueError):
        analyze_ctmc(states=states, rate_matrix=rate_matrix, t=t, init=q_init)
    with pytest.raises(ValueError):
        analyze_ctmc(states=states, rate_matrix=rate_matrix, d=d, init=q_init)
    with pytest.raises(ValueError):
        analyze_ctmc(states=states, rate_matrix=rate_matrix, n=n, init=q_init)
    with pytest.raises(ValueError):
        analyze_ctmc(states=states, rate_matrix=rate_matrix, t=3, d=1,
                     n=5, init=q_init)

    captured_output = io.StringIO()
    sys.stdout = captured_output
    print_markov(analysis, mtype="ctmc")
    sys.stdout = sys.__stdout__  # reset stdout
    test_str = ("Transition Rates:\n"
                "[2.  4.5 4.5 4.5 4.5 2.5]\n"
                "P:\n"
                "[[0.9   0.1   0.    0.    0.    0.   ]\n"
                " [0.125 0.775 0.1   0.    0.    0.   ]\n"
                " [0.    0.125 0.775 0.1   0.    0.   ]\n"
                " [0.    0.    0.125 0.775 0.1   0.   ]\n"
                " [0.    0.    0.    0.125 0.775 0.1  ]\n"
                " [0.    0.    0.    0.    0.125 0.875]]\n"
                "Steady State:\n"
                "[0.2710556  0.21684448 0.17347558 0.13878047 0.11102437"
                " 0.0888195 ]\n"
                "Initial States:\n"
                "[1, 0, 0, 0, 0, 0]\n"
                "Transient Probabilities:\n"
                "[0.43253628 0.29099199 0.16203962 0.0745227  0.02883399"
                " 0.01107543]\n"
                "Generator Matrix:\n"
                "[[ 1.   2.   0.   0.   0.   0. ]\n"
                " [ 1.  -4.5  2.   0.   0.   0. ]\n"
                " [ 1.   2.5 -4.5  2.   0.   0. ]\n"
                " [ 1.   0.   2.5 -4.5  2.   0. ]\n"
                " [ 1.   0.   0.   2.5 -4.5  2. ]\n"
                " [ 1.   0.   0.   0.   2.5 -2.5]]\n")
    assert captured_output.getvalue() == test_str



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
sys.path.insert(0, os.path.abspath('../..'))


# -- Project information -----------------------------------------------------

project = 'ORMM'
copyright = '2020, Ethan Buck'
author = 'Ethan Buck'

# The full version, including alpha/beta/rc tags
release = '0.0.1'


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinx.ext.coverage",
    "sphinx.ext.mathjax",
    "sphinx.ext.autosummary"
]
napoleon_google_docstring = False
napoleon_numpy_docstring = True
autosummary_generate = True

# list of modules to mock import
autodoc_mock_imports = ["pyomo", "quantecon"]

# overwrite default master doc - needed for readthedocs.org
master_doc = "index"

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
on_rtd = os.environ.get('READTHEDOCS', None) == 'True'

if not on_rtd:  # only import and set the theme if we're building docs locally
    import sphinx_rtd_theme
    html_theme = 'sphinx_rtd_theme'
    html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]

    def setup(app):
        """Override css default theme"""
        app.add_css_file('theme_overrides.css')

    html_context = {
        'css_files': [
            '_static/theme_overrides.css',
        ],
    }
else:
    html_context = {
        'css_files': [
            'https://media.readthedocs.org/css/sphinx_rtd_theme.css',
            'https://media.readthedocs.org/css/readthedocs-doc-embed.css',
            '_static/theme_overrides.css',
        ],
    }


# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']



import pyomo.environ as pyo
import pytest

from ormm.mathprog import scheduling
from tests.methods import solve_instance

DATA_PATH = "ormm/mathprog/example_data/"
EMPLOYEE_DATA = DATA_PATH + "employee.dat"
RENTAL_DATA = DATA_PATH + "rental.dat"
AGG_PLANNING_DATA = DATA_PATH + "aggregate_planning.dat"


def test_scheduling_error():
    with pytest.raises(TypeError):
        scheduling("not a scheduling prob class")


def test_employee_simple():
    model = scheduling("employee")
    instance1 = model.create_instance(EMPLOYEE_DATA)
    instance2 = scheduling("employee", filename=EMPLOYEE_DATA)
    for inst in [instance1, instance2]:
        instance, results = solve_instance(inst)
        assert instance.OBJ() == 215
        for v in instance.component_objects(pyo.Var, active=True):
            assert {index: v[index].value for index in v} == {
                1: 0, 2: 65, 3: 70, 4: 30, 5: 0, 6:  45, 7: 5}


def test_rental_simple():
    model = scheduling(prob_class="rental")
    instance1 = model.create_instance(RENTAL_DATA)
    instance2 = scheduling(prob_class="rental", filename=RENTAL_DATA)
    for inst in [instance1, instance2]:
        instance, results = solve_instance(inst)
        assert instance.OBJ() == 1830.0
        for v in instance.component_objects(pyo.Var, active=True):
            assert {index: v[index].value for index in v} == {
                ('Mon', 'DailyWeekDay'): 1.0,
                ('Tue', 'DailyWeekDay'): 0.0,
                ('Wed', 'DailyWeekDay'): 3.0,
                ('Thu', 'DailyWeekDay'): 0.0,
                ('Fri', 'DailyWeekDay'): 2.0,
                ('Sat', 'DailyWeekend'): 0.0,
                ('Sun', 'DailyWeekend'): 0.0,
                ('Mon', 'ThreeWeekDay'): 1.0,
                ('Tue', 'ThreeWeekDay'): 0.0,
                ('Wed', 'ThreeWeekDay'): 0.0,
                ('Sat', 'Weekend'): 0.0,
                ('Mon', 'AllWeekDay'): 4.0,
                ('Sat', 'AllWeek'): 4.0}


def test_agg_planning():
    model = scheduling(prob_class="agg_planning")
    instance1 = model.create_instance(AGG_PLANNING_DATA)
    instance2 = scheduling(prob_class="agg_planning",
                           filename=AGG_PLANNING_DATA)
    for inst in [instance1, instance2]:
        instance, results = solve_instance(inst)
        assert instance.OBJ() == 865_650
        for v in instance.component_objects(pyo.Var, active=True):
            if v.name == "Produce":
                assert {index: v[index].value
                        for index in v} == {'Jan': 1350.0, 'Feb': 1400.0,
                                            'Mar': 1000.0, 'Apr': 550.0,
                                            'May': 1700.0, 'Jun': 2000.0}
            else:
                assert {index: v[index].value
                        for index in v} == {'Jan': 250.0, 'Feb': 250.0,
                                            'Mar': 250.0, 'Apr': 0.0,
                                            'May': 0.0, 'Jun': 100.0}



"""
Eventually may move transportation_model to method in Graph class
To do this, would:
  * When call transportation, arguments for:
    - These are the supply nodes
    - These are their supply max's
        source = {"A": 35, "B": 50}
    - These are demand nodes
    - These are demand requirements (mins)
        destination = {"C": 20, "D": 25, "E": 40}
    - Then maybe option for adding that info to graph object?

Other ideas to add:
  * Implement print() method on Graph object (__str__)
  * Graph() should be able to read in data file (.dat)
    used for transportation problem
    - either when transportation_problem is called, or
      in separate method to permanently add attributes
  * Graph() needs add_arcs overhaul
    - reading in different attributes
    - different input types to handle:
      1. Dictionaries
      2. Lists
      3. Numpy arrays
      4. Pandas DF
      5. Data file (.dat)?
"""

from collections import defaultdict
from collections.abc import Sequence

import numpy as np
import pandas as pd
import pyomo.environ as pyo


def transportation_model(**kwargs):
    """
    Factory method for the balanced transportation problem.

    By balanced, we mean that this implementation currently requires the data
    to have the same number of source nodes as destination nodes. Your data
    can be easily changed to meet this requirement; see the notes section.

    This network flow problem has a set of source nodes and
    destination nodes, with shipping costs between each of them.
    There are demands at the destinations, and supply limits at the
    sources.  The objective is to minimize the shipping costs while
    meeting the demands.

    Parameters
    ----------
    **kwargs
        Passed into Pyomo Abstract Model's `create_instance`
        to return Pyomo Concrete Model instead.

    Returns
    -------
    pyomo.environ.AbstractModel or pyomo.environ.ConcreteModel
        Abstract Model with the sets, parameters, decision variables,
        objective, and constraints for the transportation problem.
        Returns a Concrete Model instead if any kwargs passed.

    Notes
    -----
    This is a bipartite network with m supply nodes and n destination nodes.
    If not possible to ship from i to j, a large cost M should be passed.

    Assumes the feasibility property holds (total supply equals total demand) -
    then becomes balanced TP. You can modify data so this requirement is
    satisfied.

    Let :math:`\\delta` be the excess amount (positive).  Add  a dummy source
    to the data with index m + 1 if demand > supply.

    :math:`s_{m+1} = \\delta\\text{ ; }c_{m+1,j}=0 \\quad \\forall j \\in J`

    Add a dummy demand to the data with index n + 1 if supply > demand.

    :math:`d_{n+1} = \\delta\\text{ ; }c_{i,n+1}=0 \\quad \\forall i \\in I`

    .. math::

        \\text{Min} \\sum_{i \\in I}\\sum_{j \\in J}C_{i,j}X_{i,j}

        \\text{s.t. } \\sum_{j \\in J} X_{i,j} = S_i \\quad \\forall i \\in I

        \\sum_{i \\in I} X_{i,j} = D_j \\quad \\forall j \\in J

        X_{i,j} \\geq 0\\text{, int} \\quad \\forall i \\in I\\text{, }j \\in J
    """
    def _obj_expression(model):
        """Objective Expression: Minimizing Shipping Costs"""
        return pyo.summation(model.ShippingCosts, model.Flows)

    def _supply_constraint_rule(model, i):
        """Constraints for flows from supply"""
        return sum(model.Flows[i, j]
                   for j in model.Destinations) == model.Supply[i]

    def _demand_constraint_rule(model, j):
        """Constraints for flows from demands"""
        return sum(model.Flows[i, j]
                   for i in model.Supply) == model.Demand[j]

    # Create the abstract model & dual suffix
    model = pyo.AbstractModel()
    model.dual = pyo.Suffix(direction=pyo.Suffix.IMPORT)
    # Define sets/params that are always used
    model.Sources = pyo.Set()
    model.Destinations = pyo.Set()
    model.Supply = pyo.Param(model.Sources)
    model.Demand = pyo.Param(model.Destinations)
    model.ShippingCosts = pyo.Param(model.Sources, model.Destinations)
    # Define decision variables
    model.Flows = pyo.Var(
        model.Sources,
        model.Destinations,
        within=pyo.NonNegativeReals,
        bounds=(0, None))
    # Define objective & constraints
    model.OBJ = pyo.Objective(rule=_obj_expression, sense=pyo.minimize)
    model.SupplyConstraint = pyo.Constraint(
        model.Sources,
        rule=_supply_constraint_rule)
    model.DemandConstraint = pyo.Constraint(
        model.Destinations,
        rule=_demand_constraint_rule)
    # Check if returning concrete or abstract model
    if kwargs:
        return model.create_instance(**kwargs)
    else:
        return model


class Graph():
    """
    Fully connected network of nodes via arcs with costs.
    This class can be used to create graphs (networks), and
    then solve different network flow models on these graphs
    such as the transportation model and shortest path model.

    Parameters
    ----------
    arcs : :py:obj:`dict`, optional
        All possible paths (arcs) from each node.
        e.g. {'A': ['B', 'C', 'D', 'E'], ...}
    costs : :py:obj:`dict`, optional
        The cost of traveling from one node to another.
        e.g. {('A', 'B'): 2, ('A', 'C'): 5, ...}
    nodes : :py:obj:`set`, optional
        A set of all unique nodes in the graph.
        e.g. {"A", "B", "C", ...}
    """
    def __init__(self, arcs=None, costs=None, nodes=None):
        self.arcs = arcs if arcs is not None else defaultdict(list)
        self.costs = costs if costs is not None else {}
        self.nodes = nodes if nodes is not None else set()

    def __repr__(self):
        """
        String representation that represents data types
        and contents for developers
        """
        return (f"Graph({self.arcs!r}, {self.costs!r}, {self.nodes!r})")

    def __str__(self):
        """
        String representation for user reading object

        NOT IMPLEMENTED - currently just calls __repr__
        """
        return repr(self)

    def add_arcs(self, arc_data):
        """
        Parameters
        ----------
        arc_data
            Iterable of Iterables that contain arc information,
            such as from node ("From"), to node ("To"),
            cost ("Cost), and
            optionally direction ("Direction") -
            whether the arc is one-directional
            ("one") or bi-directional ("two").

        Raises
        ------
        TypeError
            If `arcs` is not a dict, list, DataFrame, tuple, or array
        ValueError
            If `arcs` does not have enough arguments (columns) to support
            the requirements (from_node, to_node, cost)

        Examples
        --------
        Adding arcs via list of lists.

        >>> arcs_data = [["A", "B", 7, "one"],
        ...              ["B", "C", 3, "one"],
        ...              ["C", "A", 8, "two"]]
        >>> graph = Graph().add_arcs(arcs_data)

        Adding arcs via dictionary - keys must match.

        >>> arcs_data_dict = {"From": ["A", "B", "C"],
        ...                   "To": ["B", "C", "A"],
        ...                   "Cost": [7, 3, 8],
        ...                   "Direction": ["one", "one", "two"]}
        >>> graph = Graph().add_arcs(arcs_data_dict)

        Adding arcs via Pandas DataFrame - column names must be the same, or
        there must be exactly four columns which will be read in order.

        >>> arcs_data_df = pd.DataFrame(arcs_data,
        ...                             columns=["From", "To",
        ...                                      "Cost", "Direction"])
        >>> graph = Graph().add_arcs(arcs_data_df)
        >>> arcs_data_df = pd.DataFrame(arcs_data)
        >>> graph = Graph().add_arcs(arcs_data_df)
        """
        if isinstance(arc_data, (Sequence, np.ndarray)):
            for arc in arc_data:
                self._add_arc(*arc)
        elif isinstance(arc_data, (pd.DataFrame, dict)):
            if isinstance(arc_data, dict):
                arc_data = pd.DataFrame(arc_data)
            # Create possible keywords
            from_choices = pd.Series(["From", "FromNode", "From_Node",
                                      "From_node", "From Node"])
            to_choices = pd.Series(["To", "ToNode", "To_Node", "To_node",
                                    "To Node"])
            cost_choices = pd.Series(["Cost", "ShippingCost", "Shipping_Cost",
                                      "Shipping_cost", "Shipping Cost"])
            # Add lowercase & uppercase options to choices
            from_choices = from_choices.append(
                from_choices.str.lower(), ignore_index=True).append(
                    from_choices.str.upper(), ignore_index=True)
            to_choices = to_choices.append(
                to_choices.str.lower(), ignore_index=True).append(
                    to_choices.str.upper(), ignore_index=True)
            cost_choices = cost_choices.append(
                cost_choices.str.lower(), ignore_index=True).append(
                    cost_choices.str.upper(), ignore_index=True)
            # Get lists of matches in column names
            from_matches = [col for col in arc_data.columns
                            if col in set(from_choices)]
            to_matches = [col for col in arc_data.columns
                          if col in set(to_choices)]
            cost_matches = [col for col in arc_data.columns
                            if col in set(cost_choices)]
            if from_matches and to_matches and cost_matches:
                direction_choices = pd.Series(["Direction", "direction"])
                direction_matches = [col for col in arc_data.columns
                                     if col in set(direction_choices)]
                if direction_matches:
                    key_cols = {"From": from_matches[0],
                                "To": to_matches[0],
                                "Cost": cost_matches[0],
                                "Direction": direction_matches[0]}
                    for _, row in arc_data.iterrows():
                        self._add_arc(row[key_cols["From"]],
                                      row[key_cols["To"]],
                                      row[key_cols["Cost"]],
                                      row[key_cols["Direction"]])
                else:
                    key_cols = {"From": from_matches[0],
                                "To": to_matches[0],
                                "Cost": cost_matches[0]}
                    for _, row in arc_data.iterrows():
                        self._add_arc(row[key_cols["From"]],
                                      row[key_cols["To"]],
                                      row[key_cols["Cost"]])
            else:
                num_cols = arc_data.shape[1]
                if num_cols == 3:
                    for _, row in arc_data.iterrows():
                        self._add_arc(row[0], row[1], row[2])
                elif num_cols >= 4:
                    for _, row in arc_data.iterrows():
                        self._add_arc(row[0], row[1], row[2], row[3])
                else:
                    raise ValueError("Not enough columns in `arcs` to support"
                                     " required arguments (`From`, `To`,"
                                     " `Cost`)")
        else:
            raise TypeError("Argument 'arcs' must be an iterable of"
                            " iterables!")

    def _add_arc(self, from_node, to_node, cost, direction="two"):
        # Check if to_node in from_node's list in self.arcs dict
        if from_node not in self.arcs or to_node not in self.arcs[from_node]:
            self.arcs[from_node].append(to_node)

        # Add new cost, or overwrite old cost
        self.costs[(from_node, to_node)] = cost

        # Do same thing as above in reverse if bidirectional arc
        if direction == "two":
            if to_node not in self.arcs and (
                    from_node not in self.arcs[to_node]):
                self.arcs[to_node].append(from_node)
            self.costs[(to_node, from_node)] = cost

        # Add from_node and to_node to nodes set if don't exist
        self.nodes.update([from_node, to_node])

    def shortest_path(self, source):
        """
        Solve the shortest path tree problem with Dijkstra's Algorithm.
        This requires nonnegative costs/arc lengths to get an optimal solution.

        Parameters
        ----------
        source
            The source node to use for minimizing the distance to all
            other nodes

        Returns
        -------
        dictionary
            Contains minimum costs and best paths
            to achieve those minimum costs

        Raises
        ------
        ValueError
            If self.costs contains any negative costs (arc lengths)
        """
        if min(self.costs.values()) < 0:
            raise ValueError("Non-negative costs (arc lengths) required"
                             " for Dijkstra's Algorithm for"
                             " shortest path tree problem!")
        solved_nodes = {source}
        min_costs = {source: 0}
        best_paths = {source: (source,)}
        while solved_nodes != self.nodes:
            # Find best arc that passes from solved node to unsolved
            # Filter self.arcs to include only arcs from solved nodes
            valid_arcs = {node: self.arcs[node]
                          for node in solved_nodes if self.arcs[node]}
            # Change valid_arcs from dict of lists to list of tuples
            #  And take out arcs that go to solved_nodes
            arc_tuples = [(key, *dest) for key, value in valid_arcs.items()
                          for dest in value]
            arc_tuples = [(start, end) for (start, end) in arc_tuples
                          if end not in solved_nodes]
            # Retrieve costs for arc_tuples
            costs = {node: self.costs[node] for node in arc_tuples}
            total_costs = {(from_node, to_node):
                           min_costs[from_node] + costs[(from_node, to_node)]
                           for (from_node, to_node) in arc_tuples}
            best_add = min(total_costs, key=total_costs.get)
            best_cost = total_costs[best_add]
            new_solved_node = best_add[1]
            solved_nodes.add(new_solved_node)
            min_costs[new_solved_node] = best_cost
            best_paths[new_solved_node] = \
                best_paths[best_add[0]] + (new_solved_node,)
        return {"Costs": min_costs, "Paths": best_paths}

    def transportation(self, supply, demand):
        """
        Return Concrete Model for Balanced Transportation Problem.

        This calls :py:obj:`transportation_model()` for this Graph
        object and the arguments of this method.  For more details
        on the model, see the network module function's docstring.

        Parameters
        ----------
        supply: dict
            dict of source nodes and their supply capacities
            e.g. {"A": 15, "B": 35}
        demand: dict
            dict of destination nodes and their demand requirements
            e.g. {"C": 20, "D": 30}

        Raises
        ------
        ValueError
            A source or destination node given does not exist in
            the graph object
        """
        # Data that needs to be used for concrete model
        for source in supply.keys():
            if source not in self.nodes:
                raise ValueError(f"Source Node {source} does not " +
                                 "exist in the Graph!")
        for dest in demand.keys():
            if dest not in self.nodes:
                raise ValueError(f"Destination Node {dest} does not " +
                                 "exist in the Graph!")
        transp_data = {None: {"Sources": supply.keys(),
                              "Destinations": demand.keys(),
                              "Supply": supply,
                              "Demand": demand,
                              "ShippingCosts": dict(self.costs)}}
        instance = transportation_model(data=transp_data)
        return instance



from setuptools import setup, find_packages

LONG_DESCRIPTION = """
Operations Research Models & Methods (ORMM) is inspired by Paul A. Jensen's
Excel Add-ins.  His Excel packages were last updated in 2011, and while I
believe they do still work, his work may become outdated in a couple of ways:

- Excel is not as commonly used for OR, except in settings where security is
  of the utmost concern and/or modern languages like Python, R, Julia, C, C++,
  MATLAB, AMPL, or other modeling software are not available.
- From what I understand, Microsoft has been trying to phase out VBA and move
  to Javascript.  If this happens, this could significantly impact whether or
  not his packages will work.
- While his website and packages are still available
  `here <https://www.me.utexas.edu/~jensen/ORMM/>`_, some sections are/may
  become unusable.  The animations rely on Flash, which is being phased out
  in google chrome and other web browsers.

This python package aims to accomplish some of the same goals as Paul
Jensen's website and add-ins did, mainly to

1. Be an educational tool that shows how abstract models (linear programs,
   integer programs, nonlinear programs, etc.) can be applied to real-life
   scenarios to solve complex problems.
2. Help the practitioner by providing modeling frameworks, methods for solving
   these models, and problem classes so a user can more easily see how they
   may be able to frame their business problem/objective through the lens of
   Operations Research.

This repository contains subpackages for grouping the different types of OR
Models & Methods.  Currently this subpackage list includes

1. `mathprog`: A subpackage for mathematical programs, including linear
   programs and mixed integer linear programs.
2. `markov`: A subpackage for discrete state markov analysis.
3. `network`: A subpackage for network models and methods, including the
   transportation and shortest path tree problems.
"""

PROJECT_URLS = {
    "Documentation": "https://ormm.readthedocs.io/en/stable/",
}

setup(
    name="ormm",
    version="0.1.0",
    description="A collection of Operations Research Models & Methods",
    url="https://github.com/egbuck/ormm",
    author="Ethan Buck",
    author_email="egbuck96@gmail.com",
    packages=find_packages(include=["ormm", "ormm.*"]),
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/x-rst",
    install_requires=[
        "pyomo >= 5.0",
        "pandas >= 1",
        "quantecon >= 0.4"
    ],
    extras_require={
        "dev": [
            "pytest >= 6.0",
            "sphinx >= 3.1.2",
            "sphinx_rtd_theme >= 0.5.0",
            "twine >= 3.2.0",
            "flake8 >= 3.8.3",
            "pytest-cov >=2.10.0",
            "codecov >= 2.1.8"
            # "check-manifest>=0.42" # used for creating Manifest.in
        ]
    },
    classifiers=[
        "Programming Language :: Cython",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        ("License :: OSI Approved :: "
         "GNU General Public License v3 or later (GPLv3+)"),
        "Operating System :: OS Independent",
        "Development Status :: 2 - Pre-Alpha",
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Manufacturing",
        "Intended Audience :: End Users/Desktop",
        "Intended Audience :: Education",
        "Topic :: Education",
        "Topic :: Scientific/Engineering :: Mathematics",
    ],
    project_urls=PROJECT_URLS,
)
